@@ -39,7 +39,7 @@ the license in the LICENSE file.
 
 """
 
-__version__ = '0.2.3'
+__version__ = '0.2.4'
 __docformat__ = "restructuredtext en"
 
 from base64 import b64encode
@@ -47,16 +47,19 @@ from datetime import datetime, timedelta
 
 try: # make it python2.4 compatible!
     from hashlib import sha256
-except ImportError:
+except ImportError: # pragma: no cover
     from Crypto.Hash import SHA256 as sha256
     
 import hmac
-from lxml import objectify
+from lxml import objectify, etree
 import re
 import socket
 from time import strftime, gmtime
-from urlparse import urlsplit
-from urllib2 import quote, urlopen, HTTPError
+import urlparse
+import urllib2
+
+USER_AGENT = ('python-amazon-product-api/%s '
+    '+http://pypi.python.org/pypi/python-amazon-product-api/' % __version__)
 
 LOCALES = {
     'ca' : 'http://ecs.amazonaws.ca/onca/xml', 
@@ -79,7 +82,7 @@ class AWSError (Exception):
     def __init__(self, code, msg):
         self.code = code
         self.msg = msg
-    def __str__(self):
+    def __str__(self): # pragma: no cover
         return '%(code)s: %(msg)s' % self.__dict__
     
 class InvalidSearchIndex (Exception):
@@ -149,21 +152,48 @@ class InvalidParameterCombination (Exception):
     Your request contained a restricted parameter combination.
     """
 
-INVALID_VALUE_REG = re.compile(
-    'The value you specified for (?P<parameter>\w+) is invalid.')
+DEFAULT_ERROR_REGS = {
+    'invalid-value' : re.compile(
+        'The value you specified for (?P<parameter>\w+) is invalid.'),
 
-INVALID_PARAMETER_VALUE_REG = re.compile('(?P<value>.+?) is not a valid value '
-    'for (?P<parameter>\w+). Please change this value and retry your request.')
+    'invalid-parameter-value' : re.compile(
+        '(?P<value>.+?) is not a valid value for (?P<parameter>\w+). Please '
+        'change this value and retry your request.'),
 
-NOSIMILARITIES_REG = re.compile(
-    'There are no similar items for this ASIN: (?P<ASIN>\w+).')
+    'no-similarities' : re.compile(
+        'There are no similar items for this ASIN: (?P<ASIN>\w+).'),
 
-NOT_ENOUGH_PARAMETERS_REG = re.compile('Your request should have atleast '
-        '(?P<numer>\d+) of the following parameters: (?P<parameters>[\w ,]+).')
+    'not-enough-parameters' : re.compile(
+        'Your request should have atleast (?P<number>\d+) of the following '
+        'parameters: (?P<parameters>[\w ,]+).'),
 
-INVALID_PARAMETER_COMBINATION_REG = re.compile(
-     'Your request contained a restricted parameter combination.'
-     '\s*(?P<message>\w.*)$') # only the last bit is of interest here
+    'invalid-parameter-combination' : re.compile(
+         'Your request contained a restricted parameter combination.'
+         '\s*(?P<message>\w.*)$') # only the last bit is of interest here
+}
+
+JAPANESE_ERROR_REGS = {
+    'invalid-value' : re.compile(
+        u'(?P<parameter>\w+)\u306b\u6307\u5b9a\u3057\u305f\u5024\u306f\u7121'
+        u'\u52b9\u3067\u3059\u3002'),
+
+    'invalid-parameter-value' : re.compile(
+        u'(?P<value>.+?)\u306f\u3001(?P<parameter>\w+)\u306e\u5024\u3068\u3057'
+        u'\u3066\u7121\u52b9\u3067\u3059\u3002\u5024\u3092\u5909\u66f4\u3057'
+        u'\u3066\u304b\u3089\u3001\u518d\u5ea6\u30ea\u30af\u30a8\u30b9\u30c8'
+        u'\u3092\u5b9f\u884c\u3057\u3066\u304f\u3060\u3055\u3044\u3002'),
+
+    'no-similarities' : re.compile(
+        'There are no similar items for this ASIN: (?P<ASIN>\w+).'),
+
+    'not-enough-parameters' : re.compile(
+        u'\u6b21\u306e\u30d1\u30e9\u30e1\u30fc\u30bf\u306e\u3046\u3061\u3001'
+        u'\u6700\u4f4e1\u500b\u304c\u30ea\u30af\u30a8\u30b9\u30c8\u306b\u542b'
+        u'\u307e\u308c\u3066\u3044\u308b\u5fc5\u8981\u304c\u3042\u308a\u307e'
+        u'\u3059\uff1a(?P<parameters>.+)$'),
+
+    'invalid-parameter-combination' : re.compile('^(?P<message>.*)$'),
+}
 
 class API (object):
     
@@ -176,8 +206,8 @@ class API (object):
         AWS_KEY = '...'
         SECRET_KEY = '...'
         
-        api = ProductAdvertisingAPI(AWS_KEY, SECRET_KEY)
-        root = api.item_lookup('987311264224', IdType='ISBN', 
+        api = ProductAdvertisingAPI(AWS_KEY, SECRET_KEY, 'us')
+        root = api.item_lookup('9783836214063', IdType='ISBN', 
                     SearchIndex='Books', ResponseGroup='Reviews', ReviewPage=1)
         
         rating = root.Items.Item.CustomerReviews.AverageRating.pyval
@@ -209,8 +239,8 @@ class API (object):
     REQUESTS_PER_SECOND = 1 #: max requests per second
     TIMEOUT = 5 #: timeout in seconds
     
-    def __init__(self, access_key_id, secret_access_key, 
-                 locale='de', processor=None):
+    def __init__(self, access_key_id, secret_access_key, locale, 
+                 processor=None):
         """
         :param access_key_id: AWS access key ID.
         :param secret_key_id: AWS secret key.
@@ -221,8 +251,9 @@ class API (object):
         self.secret_key = secret_access_key
         
         try:
-            parts = urlsplit(LOCALES[locale])
+            parts = urlparse.urlsplit(LOCALES[locale])
             self.scheme, self.host, self.path = parts[:3]
+            self.locale = locale
         except KeyError:
             raise UnknownLocale(locale)
         
@@ -230,8 +261,10 @@ class API (object):
         
         self.last_call = datetime(1970, 1, 1)
         self.throttle = timedelta(seconds=1)/self.REQUESTS_PER_SECOND
+        self.debug = 0 # set to 1 if you want to see HTTP headers
         
         self.response_processor = processor
+        self._parser = None # will be set first time _parse is run 
         
     def _build_url(self, **qargs):
         """
@@ -259,7 +292,7 @@ class API (object):
         
         # create signature
         keys = sorted(qargs.keys())
-        args = '&'.join('%s=%s' % (key, quote(str(qargs[key]))) 
+        args = '&'.join('%s=%s' % (key, urllib2.quote(str(qargs[key]))) 
                         for key in keys)
         
         msg = 'GET'
@@ -267,7 +300,7 @@ class API (object):
         msg += '\n' + self.path
         msg += '\n' + args.encode('utf-8')
         
-        signature = quote(
+        signature = urllib2.quote(
             b64encode(hmac.new(self.secret_key, msg, sha256).digest()))
         
         url = '%s://%s%s?%s&Signature=%s' % (self.scheme, self.host, self.path, 
@@ -278,19 +311,44 @@ class API (object):
         """
         Calls the Amazon Product Advertising API and objectifies the response.
         """
+        request = urllib2.Request(url)
+        request.add_header('User-Agent', USER_AGENT)
+        request.add_header('Accept-encoding', 'gzip')
+
         # Be nice and wait for some time 
         # before submitting the next request
         while (datetime.now() - self.last_call) < self.throttle: 
-            pass # Wait for it!
+            pass # Wait for it! # pragma: no cover
         self.last_call = datetime.now()
         
         try:
-            return urlopen(url)
-        except HTTPError, e:
+            opener = urllib2.build_opener() 
+            handler = urllib2.HTTPHandler(debuglevel=self.debug)
+            opener = urllib2.build_opener(handler) 
+            response = opener.open(request)
+            # handle compressed data
+            # Borrowed from Mark Pilgrim's excellent introduction
+            # http://diveintopython.org/http_web_services/gzip_compression.html
+            if response.headers.get('Content-Encoding') == 'gzip':
+                import StringIO
+                import gzip
+                stream = StringIO.StringIO(response.read())
+                return gzip.GzipFile(fileobj=stream)
+            return response
+        except urllib2.HTTPError, e:
             if e.code == 503:
                 raise TooManyRequests
             # otherwise re-raise
-            raise        
+            raise # pragma: no cover
+    
+    def _reg(self, key):
+        """
+        Returns the appropriate regular expression (compiled) to parse an error
+        message depending on the current locale. 
+        """
+        if self.locale == 'jp':
+            return JAPANESE_ERROR_REGS[key]
+        return DEFAULT_ERROR_REGS[key]
     
     def _parse(self, fp):
         """
@@ -301,7 +359,28 @@ class API (object):
         if self.response_processor:
             return self.response_processor(fp)
         
-        tree = objectify.parse(fp)
+        # define custom class lookup first time this method is called which 
+        # will hopefully reduce execution time (although it's lightning fast!) 
+        if self._parser is None:
+            
+            class SelectiveClassLookup(etree.CustomElementClassLookup):
+                """
+                Lookup mechanism for XML elements to ensure that ItemIds (such 
+                as ASINs) are always StringElements and evaluated as such. 
+                Thanks to Brian Browning for pointing this out.
+                """
+                def lookup(self, node_type, document, namespace, name):
+                    if name in ('ItemId', 'ASIN'):
+                        return objectify.StringElement
+                    
+            self._parser = etree.XMLParser()
+            lookup = SelectiveClassLookup()
+            lookup.set_fallback(objectify.ObjectifyElementClassLookup())
+#            lookup.set_fallback(etree.ElementNamespaceClassLookup(
+#                                    objectify.ObjectifyElementClassLookup()))
+            self._parser.set_element_class_lookup(lookup)
+        
+        tree = objectify.parse(fp, self._parser)
         root = tree.getroot()
         
         #~ from lxml import etree
@@ -313,15 +392,21 @@ class API (object):
         
         for error in errors:
             
-            if error.Code.text == 'AWS.InvalidParameterValue':
-                m = INVALID_PARAMETER_VALUE_REG.search(error.Message.text)
+            code = error.Code.text
+            msg = error.Message.text
+            
+            if code == 'AWS.ECommerceService.NoExactMatches':
+                raise NoExactMatchesFound
+            
+            if code == 'AWS.InvalidParameterValue':
+                m = self._reg('invalid-parameter-value').search(msg)
                 raise InvalidParameterValue(m.group('parameter'), m.group('value'))
             
-            if error.Code.text == 'AWS.RestrictedParameterValueCombination':
-                m = INVALID_PARAMETER_COMBINATION_REG.search(error.Message.text)
+            if code == 'AWS.RestrictedParameterValueCombination':
+                m = self._reg('invalid-parameter-combination').search(msg)
                 raise InvalidParameterCombination(m.group('message'))
-
-            raise AWSError(error.Code.text, error.Message.text)
+            
+            raise AWSError(code, msg)
         
         #~ from lxml import etree
         #~ print etree.tostring(root, pretty_print=True)
@@ -350,14 +435,14 @@ class API (object):
         except AWSError, e:
             
             if (e.code=='AWS.InvalidEnumeratedParameter' 
-            and INVALID_VALUE_REG.search(e.msg).group('parameter')=='SearchIndex'):
+            and self._reg('invalid-value').search(e.msg).group('parameter')=='SearchIndex'):
                 raise InvalidSearchIndex(params.get('SearchIndex'))
             
             if e.code=='AWS.InvalidResponseGroup': 
                 raise InvalidResponseGroup(params.get('ResponseGroup'))
             
             # otherwise re-raise exception
-            raise
+            raise # pragma: no cover
         
     def item_search(self, search_index, **params):
         """
@@ -399,14 +484,14 @@ class API (object):
         except AWSError, e:
             
             if (e.code=='AWS.InvalidEnumeratedParameter' 
-            and INVALID_VALUE_REG.search(e.msg)):
+            and self._reg('invalid-value').search(e.msg)):
                 raise InvalidSearchIndex(search_index)
             
             if e.code=='AWS.InvalidResponseGroup': 
                 raise InvalidResponseGroup(params.get('ResponseGroup'))
             
             # otherwise re-raise exception
-            raise
+            raise # pragma: no cover
         
     def similarity_lookup(self, *ids, **params):
         """
@@ -435,7 +520,7 @@ class API (object):
         except AWSError, e:
             
             if e.code=='AWS.ECommerceService.NoSimilarities':
-                asin = NOSIMILARITIES_REG.search(e.msg).group('ASIN')
+                asin = self._reg('no-similarities').search(e.msg).group('ASIN')
                 raise NoSimilarityForASIN(asin)
             
     def list_lookup(self, list_id, list_type, **params):
@@ -479,11 +564,11 @@ class API (object):
         except AWSError, e:
             
             if (e.code=='AWS.InvalidEnumeratedParameter' 
-            and INVALID_VALUE_REG.search(e.msg).group('parameter')=='ListType'):
+            and self._reg('invalid-value').search(e.msg).group('parameter')=='ListType'):
                 raise InvalidListType(params.get('ListType'))
             
             # otherwise re-raise exception
-            raise
+            raise # pragma: no cover
     
     def list_search(self, list_type, **params):
         """
@@ -520,18 +605,18 @@ class API (object):
         except AWSError, e:
             
             if (e.code=='AWS.InvalidEnumeratedParameter' 
-            and INVALID_VALUE_REG.search(e.msg).group('parameter')=='ListType'):
+            and self._reg('invalid-value').search(e.msg).group('parameter')=='ListType'):
                 raise InvalidListType(list_type)
             
             if e.code=='AWS.MinimumParameterRequirement': 
-                p = NOT_ENOUGH_PARAMETERS_REG.search(e.msg).group('parameters')
+                p = self._reg('not-enough-parameters').search(e.msg).group('parameters')
                 raise NotEnoughParameters(p)
             
             if e.code=='AWS.ECommerceService.NoExactMatches': 
                 raise NoExactMatchesFound
             
             # otherwise re-raise exception
-            raise
+            raise # pragma: no cover
         
     def help(self, about, help_type, response_group=None, **params):
         """
@@ -567,12 +652,12 @@ class API (object):
             return self._parse(fp)
         except AWSError, e:
             
-            m = INVALID_VALUE_REG.search(e.msg)
+            m = self._reg('invalid-value').search(e.msg)
             if e.code=='AWS.InvalidEnumeratedParameter': 
                 raise ValueError(m.group('parameter'))
                         
             # otherwise re-raise exception
-            raise
+            raise # pragma: no cover
         
     def browse_node_lookup(self, browse_node_id, response_group=None, **params):
         """
@@ -621,11 +706,11 @@ class API (object):
             return self._parse(fp)
         except AWSError, e:
             
-            if e.code=='AWS.InvalidResponseGroup': 
+            if e.code=='AWS.InvalidResponseGroup':
                 raise InvalidResponseGroup(params.get('ResponseGroup'))
             
             # otherwise re-raise exception
-            raise
+            raise # pragma: no cover
         
         
 class ResultPaginator (object):
@@ -647,6 +732,8 @@ class ResultPaginator (object):
                              SearchIndex='Books', ResponseGroup='Reviews'):
             ...
     
+    .. note: All three XPath expressions have to return integer values for the 
+       pagination to work!
     """
     
     def __init__(self, counter, current_page, total_pages, total_results, 
@@ -697,28 +784,41 @@ class ResultPaginator (object):
         Get total number of paginator pages.
         """
         try:
-            return root.xpath(self.total_pages_xpath, 
-                          namespaces={'aws' : self.nspace})[0].pyval
+            node = root.xpath(self.total_pages_xpath, 
+                          namespaces={'aws' : self.nspace})[0]
+            return node.pyval
+        except AttributeError:
+            # node has no attribute pyval so it better be a number
+            return int(node)
         except IndexError:
-            return None
+            return 0
         
     def _get_current_page_numer(self, root):
         """
-        Get number of current paginator page.
+        Get number of current paginator page. If it cannot be extracted, it is 
+        probably the first.
         """
         try:
-            return root.xpath(self.current_page_xpath, 
-                          namespaces={'aws' : self.nspace})[0].pyval
+            node = root.xpath(self.current_page_xpath, 
+                          namespaces={'aws' : self.nspace})[0]
+            return node.pyval
+        except AttributeError:
+            # node has no attribute pyval so it better be a number
+            return int(node)
         except IndexError:
-            return None
+            return 1
     
     def _get_total_results(self, root):
         """
         Get total number of results.
         """
         try:
-            return root.xpath(self.total_results_xpath, 
-                          namespaces={'aws' : self.nspace})
+            node = root.xpath(self.total_results_xpath, 
+                          namespaces={'aws' : self.nspace})[0]
+            return node.pyval
+        except AttributeError:
+            # node has no attribute pyval so it better be a number
+            return int(node)
         except IndexError:
-            return None
+            return 0
 
